import os
from collections import deque
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Union, Any, BinaryIO, Tuple
from flifile.datatypes import Datatypes, getdatatype


def readheadersize(f: BinaryIO) -> int:
    """
    Bit convoluted, since you can probably just aswell read a 4kb page,
    check if it already contains {END} and trim to get the header, or expand with another 4kb.
    But using deque is fun.
    """
    queue = deque([b" ", b" ", b" ", b" ", b" "], maxlen=5)
    stop = deque([b"{", b"E", b"N", b"D", b"}"], maxlen=5)
    while True:
        byte = f.read(1)
        queue.append(byte)
        if queue == stop:
            return f.tell()


def readheader(
    file: Union[str, os.PathLike[Any]]
) -> Tuple[dict[str, Dict[str, Dict[str, str]]], int]:
    file = Path(file)
    with file.open(mode="rb") as f:
        s = readheadersize(f)
        f.seek(0)
        h = parseheader(f.read(s))
        return h, s


def parseheader(headerstring: bytes) -> Dict[str, Dict[str, Dict[str, str]]]:
    chapter = "DEFAULT"
    section = "DEFAULT"
    header: Dict[str, Dict[str, Dict[str, str]]] = {}
    for line in headerstring.split(b"\n"):
        if line.startswith(b"{"):
            chapter = line.decode("utf-8").strip()[1:-1]
        if line.startswith(b"["):
            section = line.decode("utf-8").strip()[1:-1]
        else:
            kvp = line.split(b"=")
            if len(kvp) != 2:
                continue
            if chapter not in header.keys():
                header[chapter] = {}
            if section not in header[chapter]:
                header[chapter][section] = {}
            header[chapter][section][kvp[0].decode("utf-8").strip()] = (
                kvp[1].decode("utf-8").strip()
            )
    return header


def tellversion(header: Dict[str, Dict[str, Dict[str, str]]]) -> str:
    version = ""
    if "FLIMIMAGE" in header:  # for version 1
        if "INFO" in header["FLIMIMAGE"]:
            version = header["FLIMIMAGE"]["INFO"].get("version", "")
            if version:
                return version

    if "FLIMIMAGE" in header:  # for version 2
        if "DEFAULT" in header["FLIMIMAGE"]:
            version = header["FLIMIMAGE"]["DEFAULT"].get("version", "")
            if version:
                return version
    if "DEFAULT" in header:  # for version 2 if the chapter is gone.
        if "DEFAULT" in header["DEFAULT"]:
            version = header["DEFAULT"]["DEFAULT"].get("version", "")
            if version:
                return version
    return version


@dataclass
class DataInfo:
    version: str
    IMSize: tuple[int, int, int, int, int, int, int]  # ch, x, y, z, ph, t, freq
    IMType: Datatypes
    Compression: int
    BG_present: bool
    BGSize: tuple[int, int, int, int, int, int, int]  # ch, x, y, z, ph, t, freq
    BGType: Datatypes
    valid: bool

    def __bool__(self) -> bool:
        return self.valid


def telldatainfo(header: Dict[str, Dict[str, Dict[str, str]]]) -> DataInfo:
    version = tellversion(header)
    imsize = (0, 0, 0, 0, 0, 0, 0)
    imtype = Datatypes.UINT8
    compression = 0
    bgpresent = False
    bgsize = (0, 0, 0, 0, 0, 0, 0)
    bgtype = Datatypes.UINT8
    valid = False
    if version == "1.0":
        imsize = (
            int(header["FLIMIMAGE"]["LAYOUT"]["channels"]),
            int(header["FLIMIMAGE"]["LAYOUT"]["x"]),
            int(header["FLIMIMAGE"]["LAYOUT"]["y"]),
            int(header["FLIMIMAGE"]["LAYOUT"]["z"]),
            int(header["FLIMIMAGE"]["LAYOUT"]["phases"]),
            int(header["FLIMIMAGE"]["LAYOUT"]["timestamps"]),
            int(header["FLIMIMAGE"]["LAYOUT"]["frequencies"]),
        )
        imtype = getdatatype(
            header["FLIMIMAGE"]["LAYOUT"]["datatype"],
            header["FLIMIMAGE"]["LAYOUT"]["pixelFormat"],
        )
        compression = int(header["FLIMIMAGE"]["INFO"]["compression"])
        bgpresent = bool(header["FLIMIMAGE"]["LAYOUT"].get("hasDarkImage", "0") == "1")
        bgsize = (
            1,
            imsize[1],
            imsize[2],
            1,
            1,
            1,
            1,
        )
        bgtype = imtype
        if "BACKGROUND" in header["FLIMIMAGE"]:
            bgpresent = True
            bgsize = (
                int(header["FLIMIMAGE"]["BACKGROUND"]["channels"]),
                int(header["FLIMIMAGE"]["BACKGROUND"]["x"]),
                int(header["FLIMIMAGE"]["BACKGROUND"]["y"]),
                int(header["FLIMIMAGE"]["BACKGROUND"]["z"]),
                int(header["FLIMIMAGE"]["BACKGROUND"]["phases"]),
                int(header["FLIMIMAGE"]["BACKGROUND"]["timestamps"]),
                int(header["FLIMIMAGE"]["BACKGROUND"]["frequencies"]),
            )
            bgtype = getdatatype(
                header["FLIMIMAGE"]["BACKGROUND"]["datatype"],
                header["FLIMIMAGE"]["BACKGROUND"]["pixelFormat"],
            )
        valid = True
    elif version == "2.0":
        ch = len(header["FLIMIMAGE"]["DEFAULT"]["channels"].strip("{}[]").split(","))
        ph = len(header["FLIMIMAGE"]["DEFAULT"]["phases"].strip("{}[]").split(","))
        ts = len(
            header["FLIMIMAGE"]["DEFAULT"]["timestamps"].strip("{}[]").split(",")
        )  # seems unused
        fr = len(header["FLIMIMAGE"]["DEFAULT"]["frequencies"].strip("{}[]").split(","))
        imsize = (
            ch,
            int(header["FLIMIMAGE"]["DEFAULT"]["x"]),
            int(header["FLIMIMAGE"]["DEFAULT"]["y"]),
            int(header["FLIMIMAGE"]["DEFAULT"]["z"]),
            ph,
            int(header["FLIMIMAGE"]["DEFAULT"]["numberOfFrames"]),
            fr,
        )
        imtype = getdatatype(
            "",
            header["FLIMIMAGE"]["DEFAULT"]["pixelFormat"],
        )
        compression = 0
        bgpresent = False
        bgsize = (
            1,
            imsize[1],
            imsize[2],
            1,
            1,
            1,
            1,
        )
        bgtype = imtype
        valid = True

    return DataInfo(
        version=version,
        IMSize=imsize,
        IMType=imtype,
        Compression=compression,
        BG_present=bgpresent,
        BGSize=bgsize,
        BGType=bgtype,
        valid=valid,
    )



from enum import Enum
from typing import Union
import numpy as np


class Packing(Enum):
    LSB = 0
    MSB = 1
    UNKNOWN = 3


np_dtypes = Union[
    np.uint8,
    np.uint16,
    np.uint32,
    np.int8,
    np.int16,
    np.int32,
    np.float32,
    np.float64,
]


class Datatypes(Enum):
    UINT8 = (np.uint8, 8, Packing.UNKNOWN)
    UINT16 = (np.uint16, 16, Packing.UNKNOWN)
    UINT12 = (np.uint16, 12, Packing.UNKNOWN)
    UINT32 = (np.uint32, 32, Packing.UNKNOWN)
    INT8 = (np.int8, 8, Packing.UNKNOWN)
    INT16 = (np.int16, 16, Packing.UNKNOWN)
    INT32 = (np.int32, 32, Packing.UNKNOWN)
    REAL32 = (np.float32, 32, Packing.UNKNOWN)
    REAL64 = (np.float64, 64, Packing.UNKNOWN)
    BayerBG16 = (np.uint16, 16, Packing.UNKNOWN)
    BayerGB16 = (np.uint16, 16, Packing.UNKNOWN)
    BayerRG10 = (np.uint16, 16, Packing.UNKNOWN)
    BayerRG12 = (np.uint16, 16, Packing.UNKNOWN)
    BayerRG16 = (np.uint16, 16, Packing.UNKNOWN)
    Mono10 = (np.uint16, 16, Packing.UNKNOWN)
    Mono12 = (np.uint16, 16, Packing.UNKNOWN)
    Mono14 = (np.uint16, 16, Packing.UNKNOWN)
    Mono16 = (np.uint16, 16, Packing.UNKNOWN)
    Mono8 = (np.uint8, 8, Packing.UNKNOWN)
    Mono10p = (np.uint16, 10, Packing.LSB)
    Mono10pmsb = (np.uint16, 10, Packing.MSB)
    BayerBG12 = (np.uint16, 16, Packing.UNKNOWN)
    BayerBG12p = (np.uint16, 12, Packing.LSB)
    BayerBG12pmsb = (np.uint16, 12, Packing.MSB)
    BayerGB12 = (np.uint16, 16, Packing.UNKNOWN)
    BayerGB12p = (np.uint16, 12, Packing.LSB)
    BayerGB12pmsb = (np.uint16, 12, Packing.MSB)
    BayerRG12p = (np.uint16, 12, Packing.LSB)
    BayerRG12pmsb = (np.uint16, 12, Packing.MSB)
    BayerRG12Packed = (np.uint16, 12, Packing.UNKNOWN)
    Mono12p = (np.uint16, 12, Packing.LSB)
    Mono12pmsb = (np.uint16, 12, Packing.MSB)
    Mono12Packed = (np.uint16, 12, Packing.UNKNOWN)
    Mono14p = (np.uint16, 14, Packing.LSB)
    BayerBG8 = (np.uint8, 8, Packing.UNKNOWN)
    BayerGB8 = (np.uint8, 8, Packing.UNKNOWN)
    BGR8 = (np.uint8, 8, Packing.UNKNOWN)
    BGR8Packed = (np.uint8, 8, Packing.UNKNOWN)
    RGB8 = (np.uint8, 8, Packing.UNKNOWN)
    RGB8Packed = (np.uint8, 8, Packing.UNKNOWN)

    def __init__(self, v1: np_dtypes, v2: int, v3: Packing):
        self.v1 = v1
        self.v2 = v2
        self.v3 = v3

    @property
    def nptype(self) -> np_dtypes:
        return self.v1

    @property
    def bits(self) -> int:
        return self.v2

    @property
    def packing(self) -> Packing:
        return self.v3


def getdatatype(datatype: str = "", pixelformat: str = "") -> Datatypes:
    if pixelformat == "Mono8":
        return Datatypes.Mono8
    if pixelformat == "Mono12p":
        return Datatypes.Mono12p
    if datatype == "UINT8":
        return Datatypes.UINT8
    return Datatypes.UINT8



import pytest as pytest
from pathlib import Path

from flifile.readheader import readheader, tellversion, telldatainfo
from tests.testdata.headers import (
    returnheaders,
    returnversions,
    returndatastarts,
    returndatainfos,
)


@pytest.fixture
def files():
    return [
        Path(
            Path.cwd(),
            "tests",
            "testdata",
            "FliFile1.0(1)_DEV_1AB22C01C4FA_DS_0x0_02HH6.fli",
        ),
        Path(
            Path.cwd(),
            "tests",
            "testdata",
            "FliFile2.0(1)_DEV_1AB22C01C4FA_DS_0x0_02HH6.fli",
        ),
        Path(
            Path.cwd(),
            "tests",
            "testdata",
            "FliFile1.0_DEV_1AB22C01C4FA_DS_0x0_02HH6.fli",
        ),
        Path(
            Path.cwd(),
            "tests",
            "testdata",
            "FliFile2.0_DEV_1AB22C01C4FA_DS_0x0_02HH6.fli",
        ),
    ]


def testheaders(files):
    headers = returnheaders()
    versions = returnversions()
    datastarts = returndatastarts()
    datainfos = returndatainfos()
    for file in files:
        header, ds = readheader(file)
        assert header == headers[file.name]
        assert tellversion(header) == versions[file.name]
        assert ds == datastarts[file.name]
        assert telldatainfo(header) == datainfos[file.name]



"""
Python support for the Lambert Instruments .fli file

(c) R.Harkes NKI

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
from pathlib import Path
from typing import Any, Union
import numpy as np
import zlib
import logging
from .datatypes import Packing, Datatypes, np_dtypes
from .readheader import readheader, telldatainfo


class FliFile:
    """
    Lambert Instruments .fli file
    Contains:
    - header: dictionary with all header entries
    - path: pathlib.Path to the file
    Hidden:
    - _bg: to store the background
    - _datastart: pointer to the start of the data
    - _di: dictionary with data information based on the header
    """

    def __init__(self, filepath: Union[str, os.PathLike[Any]]) -> None:
        # open file
        if isinstance(filepath, str):
            self.path = Path(filepath)
        elif isinstance(filepath, Path):
            self.path = filepath
        else:
            raise ValueError("not a valid filename")
        if self.path.suffix != ".fli":
            raise ValueError("Not a valid extension")
        self.log = logging.getLogger("flifile")
        self.header, self._datastart = readheader(self.path)
        self.datainfo = telldatainfo(self.header)
        self._bg = np.array([], dtype=self.datainfo.BGType.nptype)

    def getdata(
        self, subtractbackground: bool = True, squeeze: bool = True
    ) -> np.ndarray[Any, np.dtype[np_dtypes]]:
        """
        Returns the data from the .fli file. If squeeze is False the data is retured with these dimensions:
        frequency,time,phase,z,y,x,channel
        :param subtractbackground: Subtract the background from the image data
        :param squeeze: Return data without singleton dimensions in x,y,ph,t,z,fr,c order
        :return: numpy.ndarray
        """
        if not self.datainfo.BG_present:
            subtractbackground = False
        datasize = int(np.prod(self.datainfo.IMSize, dtype=np.uint64))
        if self.datainfo.Compression > 0:
            fid = self.path.open(mode="rb")
            fid.seek(self._datastart)
            dcmp = zlib.decompressobj(32 + zlib.MAX_WBITS)  # skip the GZIP header
            data = np.frombuffer(
                dcmp.decompress(fid.read()), dtype=self.datainfo.IMType.nptype
            )
            bg = data[datasize:]
            self._bg = bg.reshape(self.datainfo.BGSize[::-1])
            data = data[:datasize]
        else:
            data = self._get_data_from_file(
                offset=self._datastart,
                datatype=self.datainfo.IMType,
                datasize=datasize,
            )

        if self.datainfo.IMType.bits == 12:  # 12 bit per pixel packed per 2 in 3 bytes
            data = self._convert_12_bit(data, datatype=self.datainfo.IMType)
        data = data.reshape(self.datainfo.IMSize[::-1])
        if subtractbackground:
            self._bg = self.getbackground(squeeze=False)
            mask = np.where(data < self._bg)
            data = data - self._bg
            data[mask] = 0
        if squeeze:
            data = np.squeeze(data.transpose((5, 4, 2, 1, 3, 0, 6)))  # x,y,ph,t,z,fr,c

        return data

    def getbackground(
        self, squeeze: bool = True
    ) -> np.ndarray[Any, np.dtype[np_dtypes]]:
        """
        Returns the background data from the .fli file. If squeeze is False the data is retured with these dimensions:
        frequency,time,phase,z,y,x,channel
        :param squeeze: Return data without singleton dimensions in x,y,ph,t,z,fr,c order
        :return: numpy.ndarray
        """
        if not self.datainfo.BG_present:
            self.log.warning("WARNING: No background present in file")
            return np.array([])
        if self._bg.size == 0:
            data = self._bg
        else:
            if self.datainfo.Compression > 0:
                self.log.warning(
                    "WARNING: Getting background before getting data is inefficient in compressed files."
                )
                self.getdata(subtractbackground=True, squeeze=False)
                data = self._bg
            else:
                offset = (
                    self._datastart
                    + (
                        self.datainfo.IMType.bits
                        * int(np.prod(self.datainfo.IMSize, dtype=np.uint64))
                    )
                    / 8
                )
                datasize = int(np.prod(self.datainfo.BGSize, dtype=np.uint64))
                data = self._get_data_from_file(
                    offset=int(offset), datatype=self.datainfo.BGType, datasize=datasize
                )
                if (
                    self.datainfo.BGType.bits == 12
                ):  # 12 bit per pixel packed per 2 in 3 bytes
                    data = self._convert_12_bit(
                        data,
                        datatype=self.datainfo.BGType,
                    )
                data = data.reshape(self.datainfo.BGSize[::-1])
        if squeeze:
            return np.squeeze(data.transpose((5, 4, 2, 1, 3, 0, 6)))  # x,y,ph,t,z,fr,c
        else:
            return data

    def getframe(
        self,
        channel: int = 0,
        z: int = 0,
        phase: int = 0,
        timestamp: int = 0,
        frequency: int = 0,
        subtractbackground: bool = True,
        squeeze: bool = True,
    ) -> np.ndarray[Any, np.dtype[np_dtypes]]:
        """
        -NOT IMPLEMENTED-
        Get a single frame from the .fli file
        :param squeeze:
        :param subtractbackground:
        :param timestamp:
        :param channel:
        :param z:
        :param phase:
        :param frequency:
        :return:
        """
        # check input
        if channel > (self.datainfo.IMSize[0] - 1):
            self.log.warning("WARNING: Channel out of range")
            return np.array([])
        if z > (self.datainfo.IMSize[3] - 1):
            self.log.warning("WARNING: Z out of range")
            return np.array([])
        if phase > (self.datainfo.IMSize[4] - 1):
            self.log.warning("WARNING: Phase out of range")
            return np.array([])
        if timestamp > (self.datainfo.IMSize[5] - 1):
            self.log.warning("WARNING: Timestamp out of range")
            return np.array([])
        if frequency > (self.datainfo.IMSize[6] - 1):
            self.log.warning("WARNING: Frequency out of range")
            return np.array([])
        # get pointer
        return np.array([])
        # get data

    @staticmethod
    def _convert_12_bit(
        data: np.ndarray[Any, np.dtype[np_dtypes]], datatype: Datatypes
    ) -> np.ndarray[Any, np.dtype[np_dtypes]]:
        datasize = int((data.size / 3) * 2)
        byte1 = data[0::3]  # contains the 8 least-significant bits of the even pixels
        byte2 = data[
            1::3
        ]  # contains the 4 most-significant bits of the even pixels and 4 least-significant of the odd pixels
        byte3 = data[2::3]  # contains the 8 least-significant bits of the odd pixels
        data = np.zeros(datasize, dtype=datatype.nptype)
        if datatype.packing == Packing.LSB:
            data[0::2] = byte1.astype(np.uint16) + np.left_shift(
                np.left_shift(byte2, 4).astype(np.uint8).astype(np.uint16), 4
            )
            data[1::2] = np.left_shift(byte3.astype(np.uint16), 4) + np.right_shift(
                byte2, 4
            ).astype(np.uint8)
        elif datatype.packing == Packing.MSB:
            data[0::2] = np.left_shift(byte1.astype(np.uint16), 4) + np.right_shift(
                byte2, 4
            ).astype(np.uint8)
            data[1::2] = np.left_shift(
                np.left_shift(byte2, 4).astype(np.uint8).astype(np.uint16), 4
            ) + byte3.astype(np.uint16)
        else:
            raise ValueError("Data has no valid packing type")
        return data

    def _get_data_from_file(
        self, offset: int, datatype: Datatypes, datasize: int
    ) -> np.ndarray[Any, np.dtype[np_dtypes]]:
        if datatype.bits == 12:  # 12 bit per pixel packed per 2 in 3 bytes
            # e.g. if you would want to get 4 values, this is 48 bits, is 6 bytes.
            count = int(3 * (datasize / 2))
            return np.fromfile(self.path, offset=offset, dtype=np.uint8, count=count)
        return np.fromfile(
            self.path, offset=offset, dtype=datatype.nptype, count=datasize
        )

    def __str__(self) -> str:
        return self.path.name
